@@ -36,6 +36,7 @@ TAG_TABLEROW = sys.intern("tablerow")
 TAG_ENDTABLEROW = sys.intern("endtablerow")
 
 
+# pylint: disable=too-many-instance-attributes
 class TableRow(Mapping[str, object]):
     """Table row helper variables."""
 
@@ -189,6 +190,10 @@ class TablerowNode(Node):
             for i, row in enumerate(tablerow):
                 buffer.write(f'<tr class="row{i+1}">')
 
+                # Hack to mimic the odd newlines in the reference implementation.
+                if i == 0:
+                    buffer.write("\n")
+
                 for j, itm in enumerate(row):
                     tablerow.step()
                     namespace[name] = itm
@@ -197,9 +202,10 @@ class TablerowNode(Node):
                     self.block.render(context=context, buffer=buffer)
                     buffer.write("</td>")
 
-                buffer.write("</tr>")
+                # Newline as per reference implementation.
+                buffer.write("</tr>\n")
 
-        return None
+        return True
 
     async def render_to_output_async(
         self, context: Context, buffer: TextIO
@@ -225,6 +231,9 @@ class TablerowNode(Node):
             for i, row in enumerate(tablerow):
                 buffer.write(f'<tr class="row{i+1}">')
 
+                if i == 0:
+                    buffer.write("\n")
+
                 for j, itm in enumerate(row):
                     tablerow.step()
                     namespace[name] = itm
@@ -233,9 +242,9 @@ class TablerowNode(Node):
                     await self.block.render_async(context=context, buffer=buffer)
                     buffer.write("</td>")
 
-                buffer.write("</tr>")
+                buffer.write("</tr>\n")
 
-        return None
+        return True
 
 
 class TablerowTag(Tag):
