@@ -36,6 +36,7 @@ from liquid import Mode
 if TYPE_CHECKING:  # pragma: no cover
     from liquid import Environment
     from liquid.template import BoundTemplate
+    from liquid.builtin.tags.for_tag import ForLoop
 
 # Maximum number of times a context can be extended or wrapped.
 MAX_CONTEXT_DEPTH = 30
@@ -47,7 +48,7 @@ Namespace = Mapping[str, object]
 _undefined = object()
 
 
-def _getitem(obj: Union[Mapping[Any, Any], Sequence[Any]], key: Any) -> object:
+def _getitem(obj: Any, key: Any) -> Any:
     """Item getter with special methods for arrays/lists and hashes/dicts."""
     # NOTE: A runtime checkable protocol was too slow.
     if hasattr(key, "__liquid__"):
@@ -89,7 +90,7 @@ def get_item(
 ) -> Any:
     """Chained item getter."""
     try:
-        itm: Any = functools.reduce(_getitem, items, obj)  # type: ignore
+        itm: Any = functools.reduce(_getitem, items, obj)
     except (KeyError, IndexError, TypeError):
         itm = default
     return itm
@@ -231,45 +232,59 @@ class DebugUndefined(Undefined):
 class StrictUndefined(Undefined):
     """An undefined that raises an exception for everything other than ``repr``."""
 
-    __slots__ = ()
+    __slots__ = ("msg",)
+
+    def __init__(self, name: str, obj: object = _undefined, hint: Optional[str] = None):
+        super().__init__(name, obj=obj, hint=hint)
+        if self.hint:
+            self.msg = self.hint
+        else:
+            self.msg = f"'{self.name}' is undefined"
 
     def __getattribute__(self, name: str) -> Any:
-        if name in ("__repr__", "name", "obj", "hint"):
+        if name in (
+            "__repr__",
+            "name",
+            "hint",
+            "obj",
+            "hint",
+            "msg",
+        ):
             return super().__getattribute__(name)
-        raise UndefinedError(f"'{self.name}' is undefined")
+        raise UndefinedError(self.msg)
 
     def __contains__(self, item: object) -> bool:
-        raise UndefinedError(f"'{self.name}' is undefined")
+        raise UndefinedError(self.msg)
 
     def __eq__(self, other: object) -> bool:
-        raise UndefinedError(f"'{self.name}' is undefined")
+        raise UndefinedError(self.msg)
 
     def __getitem__(self, key: str) -> object:
-        raise UndefinedError(f"'{self.name}' is undefined")
+        raise UndefinedError(self.msg)
 
     def __len__(self) -> int:
-        raise UndefinedError(f"'{self.name}' is undefined")
+        raise UndefinedError(self.msg)
 
     def __iter__(self) -> Iterator[Any]:
-        raise UndefinedError(f"'{self.name}' is undefined")
+        raise UndefinedError(self.msg)
 
     def __str__(self) -> str:
-        raise UndefinedError(f"'{self.name}' is undefined")
+        raise UndefinedError(self.msg)
 
     def __repr__(self) -> str:
         return f"StrictUndefined({self.name})"
 
     def __bool__(self) -> bool:
-        raise UndefinedError(f"'{self.name}' is undefined")
+        raise UndefinedError(self.msg)
 
     def __int__(self) -> int:
-        raise UndefinedError(f"'{self.name}' is undefined")
+        raise UndefinedError(self.msg)
 
     def __hash__(self) -> int:
-        raise UndefinedError(f"'{self.name}' is undefined")
+        raise UndefinedError(self.msg)
 
     def __reversed__(self) -> Iterable[Any]:
-        raise UndefinedError(f"'{self.name}' is undefined")
+        raise UndefinedError(self.msg)
 
 
 # pylint: disable=too-many-instance-attributes redefined-builtin
@@ -280,7 +295,9 @@ class Context:
         "env",
         "locals",
         "globals",
+        "counters",
         "scope",
+        "loops",
         "_tag_namespace",
         "disabled_tags",
         "autoescape",
@@ -304,19 +321,24 @@ class Context:
         # passed down from the environment.
         self.globals = globals or {}
 
+        # A namespace for `increment` and `decrement` counters.
+        self.counters: Dict[str, int] = {}
+
         # Namespaces are searched in this order. When a context is extended, the
         # temporary namespace is pushed to the front of this chain.
-        self.scope = ReadOnlyChainMap(self.locals, self.globals, builtin)
+        self.scope = ReadOnlyChainMap(self.locals, self.globals, builtin, self.counters)
 
         # A namspace supporting stateful tags. Such as `cycle`, `increment`,
         # `decrement` and `ifchanged`.
         self._tag_namespace: Dict[str, Any] = {
             "cycles": {},
-            "counters": {},
             "ifchanged": "",
             "stopindex": {},
         }
 
+        # As stack of forloop objects. Used for populating forloop.parentloop.
+        self.loops: List[object] = []
+
         # A list of tags names that are disallowed in this context. For example,
         # partial templates rendered using the "render" tag are not allowed to
         # use "include" tags.
@@ -347,9 +369,19 @@ class Context:
         if items:
             try:
                 return functools.reduce(_getitem, items, obj)
-            except (KeyError, IndexError, TypeError):
+            except (KeyError, IndexError, TypeError) as err:
+                if isinstance(err, KeyError):
+                    hint = (
+                        f"key error: {err}, {name}{''.join([f'[{i}]' for i in items])}"
+                    )
+                else:
+                    hint = f"{err}: {name}{''.join([f'[{i}]' for i in items])}"
+
                 if default == _undefined:
-                    return self.env.undefined(name)
+                    return self.env.undefined(
+                        name=name,
+                        hint=hint,
+                    )
                 return default
 
         return obj
@@ -369,9 +401,12 @@ class Context:
             try:
                 for item in items:
                     obj = await _getitem_async(obj, item)
-            except (KeyError, IndexError, TypeError):
+            except (KeyError, IndexError, TypeError) as err:
                 if default == _undefined:
-                    return self.env.undefined(name)
+                    return self.env.undefined(
+                        name=name,
+                        hint=f"{err}: {name}{''.join([f'[{i}]' for i in items])}",
+                    )
 
         return obj
 
@@ -419,14 +454,14 @@ class Context:
 
     def increment(self, name: str) -> int:
         """Increment the named counter and return its value."""
-        val: int = self._tag_namespace["counters"].get(name, -1) + 1
-        self._tag_namespace["counters"][name] = val
+        val: int = self.counters.get(name, 0)
+        self.counters[name] = val + 1
         return val
 
     def decrement(self, name: str) -> int:
         """Decrement the named counter and return its value."""
-        val: int = self._tag_namespace["counters"].get(name, 0) - 1
-        self._tag_namespace["counters"][name] = val
+        val: int = self.counters.get(name, 0) - 1
+        self.counters[name] = val
         return val
 
     def cycle(self, group_name: str, args: Sequence[Any]) -> Iterator[Any]:
@@ -470,6 +505,23 @@ class Context:
         finally:
             self.scope.pop()
 
+    @contextmanager
+    def loop(self, namespace: Namespace, forloop: ForLoop) -> Iterator[Context]:
+        """Just like ``Context.extend``, but keeps track of ForLoop objects too."""
+        self.loops.append(forloop)
+        with self.extend(namespace) as context:
+            try:
+                yield context
+            finally:
+                self.loops.pop()
+
+    def parentloop(self) -> Union[Undefined, object]:
+        """Return the last ForLoop object from the loop stack."""
+        try:
+            return self.loops[-1]
+        except IndexError:
+            return self.env.undefined("parentloop")
+
     def copy(
         self, namespace: Namespace, disabled_tags: Optional[List[str]] = None
     ) -> Context:
