@@ -50,13 +50,13 @@ class Nil(Expression):
     __slots__ = ()
 
     def __eq__(self, other: object) -> bool:
-        return isinstance(other, Nil)
+        return other is None or isinstance(other, Nil)
 
     def __repr__(self) -> str:  # pragma: no cover
-        return "Nil()"
+        return "NIL()"
 
     def __str__(self) -> str:  # pragma: no cover
-        return "nil"
+        return ""
 
     def evaluate(self, context: Context) -> None:
         return None
@@ -147,6 +147,9 @@ class Literal(Expression, Generic[T]):
     def __str__(self) -> str:
         return repr(self.value)
 
+    def __eq__(self, other: object) -> bool:
+        return self.value == other
+
     def __hash__(self) -> int:
         return hash(self.value)
 
@@ -180,6 +183,9 @@ class StringLiteral(Literal[str]):
     def __eq__(self, other: object) -> bool:
         return isinstance(other, StringLiteral) and self.value == other.value
 
+    def __hash__(self) -> int:
+        return hash(self.value)
+
     def __repr__(self) -> str:  # pragma: no cover
         return f"StringLiteral(value='{self.value}')"
 
@@ -198,6 +204,9 @@ class IntegerLiteral(Literal[int]):
     def __eq__(self, other: object) -> bool:
         return isinstance(other, IntegerLiteral) and self.value == other.value
 
+    def __hash__(self) -> int:
+        return hash(self.value)
+
     def __repr__(self) -> str:  # pragma: no cover
         return f"IntegerLiteral(value={self.value})"
 
@@ -230,9 +239,9 @@ class RangeLiteral(Expression):
         )
 
     def __str__(self) -> str:
-        return f"({self.start}..{self.stop})"
+        return f"{self.start}..{self.stop}"
 
-    def __repr__(self) -> str:
+    def __repr__(self) -> str:  # pragma: no cover
         return f"RangeLiteral(start={self.start}, stop={self.stop})"
 
     def __hash__(self) -> int:
@@ -288,6 +297,9 @@ class IdentifierPathElement(Literal[Union[int, str]]):
     def __eq__(self, other: object) -> bool:
         return isinstance(other, IdentifierPathElement) and self.value == other.value
 
+    def __hash__(self) -> int:
+        return hash(self.value)
+
     def __repr__(self) -> str:  # pragma: no cover
         return f"IdentifierPathElement(value={self.value})"
 
@@ -321,7 +333,7 @@ class Identifier(Expression):
         return ".".join(buf)
 
     def __hash__(self) -> int:
-        return hash(str(self))
+        return hash(tuple(self.path))
 
     def evaluate(self, context: Context) -> object:
         path: List[Any] = [elem.evaluate(context) for elem in self.path]
@@ -444,20 +456,21 @@ class Filter:
 
         kwargs_str = ", ".join([f"{k}: {v}" for k, v in self.kwargs.items()])
         if kwargs_str:
-            if len(buf) == 1:
-                buf.append(f": {kwargs_str}")
-            else:
-                buf.append(f", {kwargs_str}")
+            buf.append(f", {kwargs_str}")
 
         return "".join(buf)
 
     def evaluate_args(self, context: Context) -> List[object]:
+        """Return a list of filter arguments evaluated with the given context."""
         return [arg.evaluate(context) for arg in self.args]
 
     async def evaluate_args_async(self, context: Context) -> List[object]:
+        """An async version of `evaluate_args`."""
         return [await arg.evaluate_async(context) for arg in self.args]
 
     def evaluate_kwargs(self, context: Context) -> Dict[str, object]:
+        """Return a dictionary of filter keyword arguments evaluated with the given
+        context."""
         # Shortcut for the common case. Most filters do not use named
         # parameters.
         if not self.kwargs:
@@ -465,6 +478,7 @@ class Filter:
         return {k: v.evaluate(context) for k, v in self.kwargs.items()}
 
     async def evaluate_kwargs_async(self, context: Context) -> Dict[str, object]:
+        """An async version of `evaluate_kwargs`."""
         # Shortcut for the common case. Most filters do not use named
         # parameters.
         if not self.kwargs:
@@ -529,9 +543,9 @@ class FilteredExpression(Expression):
     def evaluate(self, context: Context) -> object:
         result = self.expression.evaluate(context)
 
-        for fltr in self.filters:
+        for _filter in self.filters:
             try:
-                func = context.filter(fltr.name)
+                func = context.filter(_filter.name)
             except NoSuchFilterFunc:
                 if context.env.strict_filters:
                     raise
@@ -540,8 +554,8 @@ class FilteredExpression(Expression):
             # Any exception causes us to abort the filter chain and discard the result.
             # Nothing will be rendered.
             try:
-                args = fltr.evaluate_args(context)
-                kwargs = fltr.evaluate_kwargs(context)
+                args = _filter.evaluate_args(context)
+                kwargs = _filter.evaluate_kwargs(context)
                 result = func(result, *args, **kwargs)
             except FilterValueError:
                 # Pass over filtered expressions who's left value is not allowed.
@@ -549,16 +563,18 @@ class FilteredExpression(Expression):
             except Error:
                 raise
             except Exception as err:
-                raise Error(f"filter '{fltr.name}': unexpected error: {err}") from err
+                raise Error(
+                    f"filter '{_filter.name}': unexpected error: {err}"
+                ) from err
 
         return result
 
     async def evaluate_async(self, context: Context) -> object:
         result = await self.expression.evaluate_async(context)
 
-        for fltr in self.filters:
+        for _filter in self.filters:
             try:
-                func = context.filter(fltr.name)
+                func = context.filter(_filter.name)
             except NoSuchFilterFunc:
                 if context.env.strict_filters:
                     raise
@@ -567,8 +583,8 @@ class FilteredExpression(Expression):
             # Any exception causes us to abort the filter chain and discard the result.
             # Nothing will be rendered.
             try:
-                args = await fltr.evaluate_args_async(context)
-                kwargs = await fltr.evaluate_kwargs_async(context)
+                args = await _filter.evaluate_args_async(context)
+                kwargs = await _filter.evaluate_kwargs_async(context)
                 result = func(result, *args, **kwargs)
             except FilterValueError:
                 # Pass over filtered expressions who's left value is not allowed.
@@ -576,7 +592,9 @@ class FilteredExpression(Expression):
             except Error:
                 raise
             except Exception as err:
-                raise Error(f"filter '{fltr.name}': unexpected error: {err}") from err
+                raise Error(
+                    f"filter '{_filter.name}': unexpected error: {err}"
+                ) from err
 
         return result
 
@@ -795,6 +813,7 @@ Number = Union[int, float]
 
 
 def eval_number_expression(left: Number, operator: str, right: Number) -> bool:
+    """Return the result of comparing two numbers with the given operator."""
     if operator == "<=":
         return left <= right
     if operator == ">=":
@@ -817,7 +836,9 @@ def is_truthy(obj: Any) -> bool:
     return True
 
 
+# pylint: disable=too-many-return-statements,too-many-branches
 def compare(left: Any, operator: str, right: Any) -> bool:
+    """Return the result of a comparison operation between two objects."""
     if operator == "and":
         return is_truthy(left) and is_truthy(right)
     if operator == "or":
@@ -848,6 +869,7 @@ def compare(left: Any, operator: str, right: Any) -> bool:
     if type(left) in (int, float) and type(right) in (int, float):
         return eval_number_expression(left, operator, right)
 
+    # pylint: disable=unidiomatic-typecheck
     if type(left) != type(right):
         raise LiquidTypeError(
             f"invalid operator for types '{str(left)} {operator} {str(right)}'"
